# Copyright 2021 by Tobias Erbsland / EducateIT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
This module contains useful tools to work with this library.
"""

from pathlib import Path
from typing import Optional, Union

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey, RSAPrivateKey

from .internals import RSA_KEY_SIZE


def read_public_key(public_key: Union[Path, str, bytes]) -> RSAPublicKey:
    """
    Read the public key from a file.

    :param public_key: The public key as `Path`, string or bytes object.
    :return: The public key as object to be used with the file encryption.
    """
    if isinstance(public_key, Path):
        data = public_key.read_bytes()
    elif isinstance(public_key, str):
        data = public_key.encode('utf-8')
    elif isinstance(public_key, bytes):
        data = public_key
    else:
        raise ValueError('Public key has to be a `Path`, string or bytes object.')
    key = serialization.load_pem_public_key(data)
    if not isinstance(key, RSAPublicKey):
        raise ValueError('The data does not contain a public RSA key.')
    return key


def read_private_key(private_key: Union[Path, str, bytes], password: Optional[bytes] = None) -> RSAPrivateKey:
    """
    Read the private key from a file or string.

    :param private_key: The private key as `Path`, string or bytes object.
    :param password: An optional password to decrypt the key.
    :return: The private key usable for the file decription.
    """
    if isinstance(private_key, Path):
        data = private_key.read_bytes()
    elif isinstance(private_key, str):
        data = private_key.encode('utf-8')
    elif isinstance(private_key, bytes):
        data = private_key
    else:
        raise ValueError('Private key has to be a `Path`, string or bytes object.')
    key = serialization.load_pem_private_key(data, password=password)
    if not isinstance(key, RSAPrivateKey):
        raise ValueError('The data does not contain a private RSA key.')
    return key


def save_key_pair(*, public_key: Path, private_key: Path):
    """
    This method generates a new RSA 4096 key pair and stores the private and public key in two separate
    PEM encoded files.

    Warning! The private key is stored unencrypted and should be stored somewhere secure.

    :param public_key: The path to the public key file.
    :param private_key: The path to the private key file.
    """
    if not public_key:
        raise ValueError('Missing public key path')
    if not private_key:
        raise ValueError('Missing private key path')
    if not isinstance(public_key, Path):
        raise ValueError('`public_key` has to be a `Path` object.')
    if not isinstance(private_key, Path):
        raise ValueError('`private_key` has to be a `Path` object.')
    key = rsa.generate_private_key(public_exponent=65537, key_size=RSA_KEY_SIZE)
    private_key_data = key.private_bytes(encoding=serialization.Encoding.PEM,
                                         format=serialization.PrivateFormat.PKCS8,
                                         encryption_algorithm=serialization.NoEncryption())
    private_key.write_bytes(private_key_data)
    public_key_data = key.public_key().public_bytes(encoding=serialization.Encoding.PEM,
                                                    format=serialization.PublicFormat.SubjectPublicKeyInfo)
    public_key.write_bytes(public_key_data)



# Copyright 2021 by Tobias Erbsland / EducateIT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class IntegrityError(Exception):
    """
    This exception is thrown if there is any integrity problem with the encrypted file.

    - Wrong file magic.
    - File to short.
    - Checksum mismatch.
    - Corrupt data.
    """
    pass


class DataTooLargeError(Exception):
    """
    This exception is thrown, if you set a maximum size and it would be exceeded.
    """
    pass



# Copyright 2021 by Tobias Erbsland / EducateIT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import hashlib
import io
import json
from pathlib import Path
from typing import Dict, Any, Optional, Tuple

from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric.padding import OAEP, MGF1
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey
from cryptography.hazmat.primitives.ciphers import algorithms, modes, Cipher
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat

from .errors import IntegrityError, DataTooLargeError
from .internals import AES_BLOCK_SIZE_BYTES, \
    FILE_CONFIG_TEXT, FILE_MAGIC, KNOWN_BLOCK_TYPES, FILE_SIZE_LIMIT, WORKING_BLOCK_SIZE, AES_IV_LENGTH_BYTES, \
    SIZE_ENDIANNESS, SIZE_VALUE_LENGTH, AES_KEY_LENGTH_BYTES, ENCRYPTION_DATA_GAIN, CHUNK_SIZE_LENGTH, \
    CHUNKED_BLOCK_SIZE_VALUE, MAXIMUM_BLOCK_SIZE_VALUE


class Decryptor:
    """
    The encryptor provides methods to encrypt files.
    """

    def __init__(self, private_key: RSAPrivateKey):
        """
        Create a new decryptor object.

        :param private_key: The private key to use for decryption.
        """
        self.private_key = private_key  # The public key for the encryption.
        self.source_file_handle: Optional[io.BufferedIOBase] = None  # The current source file handle.
        self.algorithm: Optional[algorithms.CipherAlgorithm] = None  # The encryption algorithm which is used.
        # Generate the hash for the given public key of the private key.
        self.public_key_hash = hashlib.sha3_512(self.private_key.public_key().public_bytes(
            encoding=Encoding.DER, format=PublicFormat.SubjectPublicKeyInfo)).digest()
        # Cache for chunked reads
        self._current_chunk_size = 0

    def _verify_magic(self):
        """
        Verify the magic of the file.
        """
        file_magic = self.source_file_handle.read(len(FILE_MAGIC))
        if file_magic != FILE_MAGIC:
            raise IntegrityError('File magic does not match.')

    def _read_block_type(self) -> bytes:
        """
        Read the block type.

        :return: The block type.
        """
        block_type = self.source_file_handle.read(4)
        if len(block_type) < 4:
            raise IntegrityError('File is not complete.')
        if block_type not in KNOWN_BLOCK_TYPES:
            raise IntegrityError('Read unknown block type.')
        return block_type

    def _read_block_size(self) -> int:
        """
        Read the block size.

        :raises IntegrityError: If there is not enough data for the size.
        :return: The block size.
        """
        size_data = self.source_file_handle.read(SIZE_VALUE_LENGTH)
        if len(size_data) < SIZE_VALUE_LENGTH:
            raise IntegrityError('File is not complete.')
        return int.from_bytes(size_data, byteorder=SIZE_ENDIANNESS, signed=False)

    def _read_chunk_size(self) -> int:
        """
        Read and expect a chunk size.

        :raises IntegrityError: If there is not enough data for the size.
        :return: The chunk size.
        """
        size_data = self.source_file_handle.read(CHUNK_SIZE_LENGTH)
        if len(size_data) < CHUNK_SIZE_LENGTH:
            raise IntegrityError('File is not complete.')
        return int.from_bytes(size_data, byteorder=SIZE_ENDIANNESS, signed=False)

    def _read_iv(self) -> bytes:
        """
        Read an IV from the file.

        :return: The IV
        """
        iv = self.source_file_handle.read(AES_IV_LENGTH_BYTES)
        if len(iv) != AES_IV_LENGTH_BYTES:
            raise IntegrityError('File is not complete.')
        return iv

    def _start_reading_chunked_data(self):
        """
        Start reading chunked data from the current source.
        """
        self._current_chunk_size = self._read_chunk_size()

    def _read_chunked_data(self, maximum_bytes_to_read: int):
        """
        Read data from a chunked section of the stream.

        :param maximum_bytes_to_read: The maximum number of bytes to read.
        :return: A block with bytes, or empty bytes block if the end of the chunked section is reached.
        """
        if maximum_bytes_to_read <= 0:
            raise ValueError('Maximum bytes to read must be greater than zero.')
        if self._current_chunk_size == 0:  # Set the block to zero if we reached the end.
            return b''
        if maximum_bytes_to_read < self._current_chunk_size:
            block = self.source_file_handle.read(maximum_bytes_to_read)
            if len(block) < maximum_bytes_to_read:
                raise IntegrityError('File is not complete.')
            self._current_chunk_size -= len(block)
        else:
            block = bytearray(self.source_file_handle.read(self._current_chunk_size))
            if len(block) < self._current_chunk_size:
                raise IntegrityError('File is not complete.')
            self._current_chunk_size = self._read_chunk_size()
            maximum_bytes_to_read -= len(block)
            while self._current_chunk_size > 0 and maximum_bytes_to_read > 0:
                bytes_to_read = min(self._current_chunk_size, maximum_bytes_to_read)
                data = self.source_file_handle.read(bytes_to_read)
                if len(data) < bytes_to_read:
                    raise IntegrityError('File is not complete.')
                self._current_chunk_size -= bytes_to_read
                if self._current_chunk_size == 0:
                    self._current_chunk_size = self._read_chunk_size()
                maximum_bytes_to_read -= len(data)
                block.extend(data)
        return block

    def _read_chunked_block_data(self, maximum_size: int, user_limit: bool) -> bytes:
        """
        Read a block with chunked data.

        :param maximum_size: The maximum size for the data
        :param user_limit: If the limit is set by the user, which causes a `DataTooLargeError` exception.
        :return: The read data.
        """
        data = bytearray()
        while chunk_size := self._read_chunk_size():
            if len(data) + chunk_size > maximum_size:
                if user_limit:
                    raise DataTooLargeError('The data exceeds the requested limit.')
                raise IntegrityError('A block exceeds the size limit.')
            data_chunk = self.source_file_handle.read(chunk_size)
            if len(data_chunk) < chunk_size:
                raise IntegrityError('File is not complete.')
            data.extend(data_chunk)
        return data

    def _read_static_block_data(self, block_size: int, maximum_size: int, user_limit: bool) -> bytes:
        """
        Read a static data block.

        :param block_size: The read block size.
        :param maximum_size: The maximum size for the data.
        :param user_limit: If the limit is set by the user, which causes a `DataTooLargeError` exception.
        :return: The read data.
        """
        if block_size > maximum_size:
            if user_limit:
                raise DataTooLargeError('The data exceeds the requested limit.')
            raise IntegrityError('A block exceeds the size limit.')
        block_data = self.source_file_handle.read(block_size)
        if len(block_data) < block_size:
            raise IntegrityError('File is not complete.')
        return block_data

    def _read_block_header(self) -> Tuple[bytes, int]:
        """
        Read the header of the next block.

        :return: The block type and the block size.
        """
        block_type = self._read_block_type()
        block_size = self._read_block_size()
        return block_type, block_size

    def _read_block(self, maximum_size: int = 100_000, user_limit: bool = False,
                    allow_chunked: bool = False) -> Tuple[bytes, bytes]:
        """
        Read the next data block in the file.

        :param maximum_size: The maximum size for this block.
        :param user_limit: If the limit is set by the user, which causes a `DataTooLargeError` exception.
        :param allow_chunked: If chunked data is allowed when reading this block.
        :return: The block type, the data in the block.
        """
        block_type, block_size = self._read_block_header()
        chunked_data = False
        if block_size == 0:
            return block_type, b''
        if block_size == CHUNKED_BLOCK_SIZE_VALUE:
            if not allow_chunked:
                raise IntegrityError('A block has chunked data where no chunked data is allowed.')
            chunked_data = True
            block_type += b'c'  # Mark the block type as chunked.
        elif block_size >= MAXIMUM_BLOCK_SIZE_VALUE:
            raise IntegrityError('A block has an invalid block size.')
        if chunked_data:
            block_data = self._read_chunked_block_data(maximum_size, user_limit)
        else:
            block_data = self._read_static_block_data(block_size, maximum_size, user_limit)
        return block_type, block_data

    def _verify_configuration(self):
        """
        Read the file configuration and verify it.
        """
        block_type, block_data = self._read_block(maximum_size=128)
        if block_type != b'CONF':
            raise IntegrityError('Expecting `CONF` block, but found another.')
        if block_data != FILE_CONFIG_TEXT:
            raise IntegrityError('File configuration does not match expectations.')

    def _verify_public_key_hash(self):
        """
        Read and verify the public key hash.
        """
        block_type, block_data = self._read_block(maximum_size=64)
        if block_type != b'EPUB':
            raise IntegrityError('Expecting `EPUB` block, but found another.')
        if block_data != self.public_key_hash:
            raise IntegrityError('The key used to encrypt this file does not match the decryption key.')

    def _read_and_decrypt_key(self):
        """
        Read and decrypt the symmetric AES key.
        """
        block_type, block_data = self._read_block(maximum_size=1000)
        if block_type != b'ESYM':
            raise IntegrityError('Expecting `ESYM` block, but found another.')
        try:
            encryption_key = self.private_key.decrypt(
                block_data,
                OAEP(mgf=MGF1(algorithm=hashes.SHA256()),
                     algorithm=hashes.SHA256(),
                     label=None))
        except ValueError as e:
            raise IntegrityError(f'Failed to read the symmetric key from the file. Error: {e}')
        if len(encryption_key) != AES_KEY_LENGTH_BYTES:
            raise IntegrityError('The decrypted key has an unexpected length.')
        self.algorithm = algorithms.AES(encryption_key)

    def _read_encrypted_block(self, expected_type: bytes, maximum_size: int = 100_000,
                              user_limit: bool = False) -> bytes:
        """
        Read and decrypt a block of data.

        :param expected_type: The expected type for the block.
        :param maximum_size: The maximum size of the decrypted data in the block.
        :param user_limit: If the maximum size is a limit set by the user.
        :return: The decrypted data from the block.
        """
        allow_chunked = (expected_type == b'DATA')
        block_type, block_data = self._read_block(maximum_size=maximum_size + ENCRYPTION_DATA_GAIN,
                                                  user_limit=user_limit,
                                                  allow_chunked=allow_chunked)
        is_chunked = False
        if block_type.endswith(b'c'):
            block_type = block_type[:4]
            is_chunked = True
        if block_type != expected_type:
            raise IntegrityError(f'Expecting `{expected_type.decode("utf-8")}` block, but found another.')
        if len(block_data) == 0:
            return b''
        if not is_chunked:
            if len(block_data) <= (SIZE_VALUE_LENGTH + AES_IV_LENGTH_BYTES):
                raise IntegrityError('The encrypted data is too short to be valid.')
            decrypted_size = int.from_bytes(block_data[:SIZE_VALUE_LENGTH], byteorder=SIZE_ENDIANNESS, signed=False)
            encrypted_size = len(block_data) - SIZE_VALUE_LENGTH - AES_IV_LENGTH_BYTES
            if decrypted_size > encrypted_size:
                raise IntegrityError(f'Decrypted size ({decrypted_size}) is larger than the actual block '
                                     f'size ({encrypted_size}).')
            if decrypted_size < 1:
                raise IntegrityError('The decrypted size is zero, which is not allowed at this point.')
            if encrypted_size % AES_BLOCK_SIZE_BYTES != 0:
                raise IntegrityError('The encrypted data does not align with the AES block size.')
            iv = block_data[SIZE_VALUE_LENGTH:AES_IV_LENGTH_BYTES + SIZE_VALUE_LENGTH]
            cipher = Cipher(self.algorithm, modes.CBC(iv))  # Create the cipher context
            cipher_context = cipher.decryptor()
            data = cipher_context.update(block_data[AES_IV_LENGTH_BYTES + SIZE_VALUE_LENGTH:])
            data += cipher_context.finalize()
        else:  # Decrypt the chunked format.
            if len(block_data) <= AES_IV_LENGTH_BYTES:
                raise IntegrityError('The encrypted data is too short to be valid.')
            iv = block_data[:AES_IV_LENGTH_BYTES]
            encrypted_size = len(block_data) - AES_IV_LENGTH_BYTES
            if encrypted_size % AES_BLOCK_SIZE_BYTES != 0:
                raise IntegrityError('The encrypted data does not align with the AES block size.')
            cipher = Cipher(self.algorithm, modes.CBC(iv))  # Create the cipher context
            cipher_context = cipher.decryptor()
            data = cipher_context.update(block_data[AES_IV_LENGTH_BYTES:])
            data += cipher_context.finalize()
            # Check the padding.
            try:
                index = data.rindex(b'\x80')
            except ValueError:
                raise IntegrityError('Invalid padding of the encrypted data: Missing padding mark.')
            if len(data) - index > AES_BLOCK_SIZE_BYTES:
                raise IntegrityError('Invalid padding of the encrypted data: Padding too large.')
            decrypted_size = index
        return data[:decrypted_size]

    def _read_and_verify_file_header(self):
        """
        Read and verify the initial portion of the file and read the encryption key.
        """
        self._verify_magic()
        self._verify_configuration()
        self._verify_public_key_hash()
        self._read_and_decrypt_key()

    def _read_metadata_block(self) -> Dict[str, Any]:
        """
        Read the metadata block and verify the checksum.
        """
        metadata_raw = self._read_encrypted_block(b'META', maximum_size=10_000)
        metadata_digest = self._read_encrypted_block(b'MDHA', maximum_size=1000)
        if len(metadata_raw) == 0:
            if len(metadata_digest) != 0:
                raise IntegrityError('The digest of the metadata block does not match.')
            return {}
        metadata_raw_digest = hashlib.sha3_512(metadata_raw).digest()
        if metadata_digest != metadata_raw_digest:
            raise IntegrityError('The digest of the metadata block does not match.')
        metadata = json.loads(metadata_raw)
        if not isinstance(metadata, dict):
            raise IntegrityError('The received metadata block was no object.')
        return metadata

    def _skip_block(self, expected_type: bytes):
        """
        Skip a block

        :param expected_type: The block type to skip.
        """
        block_type, _ = self._read_block()
        if block_type != expected_type:
            raise IntegrityError(f'Expecting `{expected_type.decode("utf-8")}` block, but found another.')

    def read_metadata(self, source: Path) -> Dict[str, Any]:
        """
        Only decrypt and read the metadata from a file.

        :param source: The file to read the metadata.
        :return: The dictionary with the metadata.
        """
        if (file_size := source.stat().st_size) < 256:
            raise IntegrityError(f'File is too short to be valid. (size={file_size})')
        with source.open('rb') as source_file_handle:
            self.source_file_handle = source_file_handle
            self._read_and_verify_file_header()
            metadata = self._read_metadata_block()
            # Stop at this point and ignore the rest of the file.
        return metadata

    def load_decrypted(self, source: Path, maximum_size: int = 10_000_000) -> bytes:
        """
        Load and decrypt the given source file.

        :param source: Load and decrypt the given source file.
        :param maximum_size: The maximum size of the decrypted data.
            This is no exact limit, because it is tested using the size of the encrypted data.
            The returned data may be up to 127 bytes larger than the given limit.
        :return: The decrypted data.
        :raises DataTooLargeError: If the maximum size would be exceeded.
        :raises IntegrityError: On any file integrity problem.
        """
        if (file_size := source.stat().st_size) < 256:
            raise IntegrityError(f'File is too short to be valid. (size={file_size})')
        with source.open('rb') as source_file_handle:
            self.source_file_handle = source_file_handle
            self._read_and_verify_file_header()
            self._skip_block(b'META')
            self._skip_block(b'MDHA')
            decrypted_data = self._read_encrypted_block(b'DATA',
                                                        maximum_size=maximum_size + AES_BLOCK_SIZE_BYTES,
                                                        user_limit=True)
            file_digest = self._read_encrypted_block(b'DTHA')
        if not decrypted_data:  # Zero file?
            if file_digest:
                raise IntegrityError('The digest of the data block does not match.')
            return b''
        decrypted_data_digest = hashlib.sha3_512(decrypted_data).digest()
        if file_digest != decrypted_data_digest:
            raise IntegrityError('The digest of the data block does not match.')
        return decrypted_data

    def _decrypt_chunked_block(self, destination: io.BufferedIOBase) -> bytes:
        """
        Read an encrypted chunked block from the current source.

        :param destination: The destination stream to write the decrypted data.
        :return: The digest of the decrypted data.
        """
        block_hash_context = hashlib.sha3_512()
        total_size = 0
        self._start_reading_chunked_data()
        iv = self._read_chunked_data(AES_IV_LENGTH_BYTES)
        if len(iv) != AES_IV_LENGTH_BYTES:
            raise IntegrityError('File is not complete.')
        cipher = Cipher(self.algorithm, modes.CBC(iv))  # Create the cipher context
        cipher_context = cipher.decryptor()
        decrypted_data = b''
        while encrypted_data := self._read_chunked_data(WORKING_BLOCK_SIZE):
            # Delay writing of the decrypted data, because we need to handle the padding at the end of the file.
            if decrypted_data:
                destination.write(decrypted_data)
                block_hash_context.update(decrypted_data)
                total_size += len(decrypted_data)
            # The encrypted data must be aligned the with the working block size.
            if len(encrypted_data) % AES_BLOCK_SIZE_BYTES != 0:
                raise IntegrityError('File is not complete: Misaligned encrypted data.')
            decrypted_data = cipher_context.update(encrypted_data)
        decrypted_data += cipher_context.finalize()  # This should not add any additional data.
        # At this point, we have the last decrypted block.
        if not decrypted_data:
            raise IntegrityError('Unexpected end of encrypted data stream.')
        if len(decrypted_data) % AES_BLOCK_SIZE_BYTES != 0:
            raise IntegrityError('Decrypted data does not align with the encryption block size.')
        try:
            index = decrypted_data.rindex(b'\x80')
        except ValueError:
            raise IntegrityError('Invalid padding of the encrypted stream: Missing padding mark.')
        if len(decrypted_data) - index > AES_BLOCK_SIZE_BYTES:
            raise IntegrityError('Invalid padding of the encrypted stream: Padding too large.')
        decrypted_data = decrypted_data[:index]  # crop the last block.
        if decrypted_data:
            destination.write(decrypted_data)
            block_hash_context.update(decrypted_data)
            total_size += len(decrypted_data)
        return block_hash_context.digest()

    def _decrypt_static_block(self, encrypted_block_size: int, destination: io.BufferedIOBase) -> bytes:
        """
        Read an encrypted static block from the current source.

        :param encrypted_block_size: The size of the encrypted block.
        :param destination: The destination stream to write the decrypted data.
        :return: The digest of the decrypted data.
        """
        if encrypted_block_size > (FILE_SIZE_LIMIT + 1000):  # > 10TB?
            raise IntegrityError('The file size is larger than 10TB, which is not supported.')
        if encrypted_block_size == 0:
            return b''  # Empty data requires an empty digest.
        block_hash_context = hashlib.sha3_512()
        encrypted_size = encrypted_block_size - SIZE_VALUE_LENGTH - AES_IV_LENGTH_BYTES
        decrypted_size = self._read_block_size()
        if decrypted_size > encrypted_size:
            raise IntegrityError('The decrypted data size is larger than the encrypted data.')
        iv = self._read_iv()
        cipher = Cipher(self.algorithm, modes.CBC(iv))  # Create the cipher context
        cipher_context = cipher.decryptor()
        encrypted_data_left = encrypted_size
        decrypted_data_left = decrypted_size
        while encrypted_data_left > 0:
            data_to_read = min(WORKING_BLOCK_SIZE, encrypted_data_left)
            encrypted_data = self.source_file_handle.read(data_to_read)
            if len(encrypted_data) < data_to_read:
                raise IntegrityError('File is not complete.')
            decrypted_data = cipher_context.update(encrypted_data)
            if decrypted_data_left < len(decrypted_data):
                decrypted_data = decrypted_data[:decrypted_data_left]
            destination.write(decrypted_data)
            block_hash_context.update(decrypted_data)
            encrypted_data_left -= len(encrypted_data)
            decrypted_data_left -= len(decrypted_data)
        # For the AES/CBC cipher, there is usually no data generated with finalize, because the data
        # is decrypted block wise. The following code is in place, in case the cipher is changed.
        decrypted_data = cipher_context.finalize()
        if decrypted_data and decrypted_data_left > 0:
            decrypted_data = decrypted_data[:decrypted_data_left]
            destination.write(decrypted_data)
            block_hash_context.update(decrypted_data)
        # At this point the decrypted data has its final length. Return the digest.
        return block_hash_context.digest()

    def _decrypt_stream(self, source_io: io.BufferedIOBase, destination_io: io.BufferedIOBase):
        """
        Decrypt a file from a source stream and write the decrypted data into the target stream.

        :param source_io: The source stream to use for decryption.
        :param destination_io: The destination stream to write the decrypted data into
        """
        self.source_file_handle = source_io
        self._read_and_verify_file_header()
        self._skip_block(b'META')
        self._skip_block(b'MDHA')
        block_type, encrypted_block_size = self._read_block_header()
        if block_type != b'DATA':
            raise IntegrityError('Expected `DATA` block, but found another one.')
        if encrypted_block_size == CHUNKED_BLOCK_SIZE_VALUE:
            data_digest = self._decrypt_chunked_block(destination_io)
        elif encrypted_block_size >= MAXIMUM_BLOCK_SIZE_VALUE:
            raise IntegrityError('The block has an invalid size value.')
        else:
            data_digest = self._decrypt_static_block(encrypted_block_size, destination_io)
        file_digest = self._read_encrypted_block(b'DTHA')
        if file_digest != data_digest:
            raise IntegrityError('The digest of the data block does not match.')

    def copy_decrypted(self, source: Path, destination: Path):
        """
        Copy a encrypted file decrypted at the destination.

        :param source: The path to the encrypted source file.
        :param destination: The path to the unencrypted destination file.
        :raises IntegrityError: On any file integrity problem.
        """
        if not source:
            raise ValueError('Missing parameter `source`')
        if not destination:
            raise ValueError('Missing parameter `destination`')
        if not isinstance(source, Path):
            raise ValueError('Parameter `source` has to be a `Path` from `pathlib`.')
        if not isinstance(destination, Path):
            raise ValueError('Parameter `destination` has to be a `Path` from `pathlib`.')
        if (file_size := source.stat().st_size) < 256:
            raise IntegrityError(f'File is too short to be valid. (size={file_size})')
        try:
            with source.open('rb') as source_io, destination.open('wb') as destination_io:
                self._decrypt_stream(source_io, destination_io)
        except IntegrityError:
            destination.unlink(missing_ok=True)
            raise

    def stream_decrypted(self, source_io: io.BufferedIOBase, destination_io: io.BufferedIOBase):
        """
        Decrypt the data from the source stream and write it to the destination stream.

        Both streams have to be open and need to be readable/writable. The implementation only
        uses the `read` method on the source stream and the `write` method on the destination stream.

        :param source_io: The open source stream.
        :param destination_io: The open destination stream.
        """
        if not source_io:
            raise ValueError('Missing parameter `source_io`')
        if not destination_io:
            raise ValueError('Missing parameter `destination_io`')
        if not isinstance(source_io, io.BufferedIOBase):
            raise ValueError('The parameter `source_io` has to be a subclass of `io.BufferedIOBase`')
        if not isinstance(destination_io, io.BufferedIOBase):
            raise ValueError('The parameter `destination_io` has to be a subclass of `io.BufferedIOBase`')
        self._decrypt_stream(source_io, destination_io)



# Copyright 2021 by Tobias Erbsland / EducateIT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import hashlib
import io
import json
import os
from datetime import datetime
from pathlib import Path
from typing import Dict, Any, Optional, Tuple

from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric.padding import OAEP, MGF1
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
from cryptography.hazmat.primitives.ciphers import algorithms, modes, Cipher, CipherContext
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat

from .errors import DataTooLargeError
from .internals import AES_BLOCK_SIZE_BYTES, \
    FILE_CONFIG_TEXT, FILE_MAGIC, KNOWN_BLOCK_TYPES, FILE_SIZE_LIMIT, WORKING_BLOCK_SIZE, SIZE_ENDIANNESS, \
    SIZE_VALUE_LENGTH, AES_KEY_LENGTH_BYTES, CHUNK_SIZE_LENGTH, MAXIMUM_CHUNK_SIZE, CHUNKED_BLOCK_SIZE_VALUE


class Encryptor:
    """
    The encryptor provides methods to encrypt files.
    """

    def __init__(self, public_key: RSAPublicKey):
        """
        Create a new encryptor object.

        :param public_key: The public key to use for encryption.
        """
        self.public_key = public_key  # The public key for the encryption.
        self.destination_file_digest = None  # The overall file digest to use.
        self.destination_file_handle: Optional[io.BufferedIOBase] = None  # The current target file handle.
        self.algorithm: Optional[algorithms.CipherAlgorithm] = None  # The encryption algorithm which is used.
        # Generate the hash for the given public key.
        self.public_key_hash = hashlib.sha3_512(self.public_key.public_bytes(
            encoding=Encoding.DER, format=PublicFormat.SubjectPublicKeyInfo)).digest()
        self._chunked_buffer = bytearray()  # A buffer to write chunked data more efficient.

    def _write_with_digest(self, data: bytes):
        """
        Write to the destination file and add to the digest.

        :param data: The data block to write.
        """
        self.destination_file_handle.write(data)
        if self.destination_file_digest:
            self.destination_file_digest.update(data)

    def _write_magic(self):
        """
        Write the magic header to the file.
        """
        self._write_with_digest(FILE_MAGIC)

    def _write_block(self, block_type: bytes, data: bytes):
        """
        Write a data block to the file handle.

        :param block_type: The block type.
        :param data: The data for the block.
        """
        if len(block_type) != 4:
            raise ValueError('Block type must be 4 bytes')
        if len(data) > 100_000:
            raise ValueError('Block data exceeds 100k')
        if block_type not in KNOWN_BLOCK_TYPES:
            raise ValueError('Block type is not known.')
        self._write_with_digest(block_type)
        size_data = len(data).to_bytes(SIZE_VALUE_LENGTH, byteorder=SIZE_ENDIANNESS, signed=False)
        self._write_with_digest(size_data)
        self._write_with_digest(data)

    def _write_chunked_block_header(self, block_type: bytes):
        """
        Write the header for a chunked block.

        :param block_type: The block type.
        """
        if len(block_type) != 4:
            raise ValueError('Block type must be 4 bytes')
        if block_type != b'DATA':
            raise ValueError('Only `DATA` blocks can use the chunked data format.')
        self._write_with_digest(block_type)
        size_data = CHUNKED_BLOCK_SIZE_VALUE.to_bytes(SIZE_VALUE_LENGTH, byteorder=SIZE_ENDIANNESS, signed=False)
        self._write_with_digest(size_data)

    def _write_configuration(self):
        """
        Write the block with the encryption configuration.
        """
        self._write_block(b'CONF', FILE_CONFIG_TEXT)

    def _write_public_key_hash(self):
        """
        Write the hash of the public key
        """
        self._write_block(b'EPUB', self.public_key_hash)

    def _create_and_write_encryption_key(self):
        """
        Create the encryption key and encrypt it.
        """
        # Generate the random encryption key for this file.
        encryption_key = os.urandom(AES_KEY_LENGTH_BYTES)
        # Encrypt the encryption key and store it in the file.
        encrypted_encryption_key = self.public_key.encrypt(
            encryption_key,
            OAEP(
                mgf=MGF1(algorithm=hashes.SHA256()),
                algorithm=hashes.SHA256(),
                label=None))
        # Write the encrypted encryption key.
        self._write_block(b'ESYM', encrypted_encryption_key)
        self.algorithm = algorithms.AES(encryption_key)

    def _prepare_encryption(self) -> Tuple[CipherContext, bytes]:
        """
        Prepare the encryption.

        Creates a new cipher context and a random initialization vector.

        :return: The cipher context, IV
        """
        iv = os.urandom(AES_BLOCK_SIZE_BYTES)  # Create a random IV
        cipher = Cipher(self.algorithm, modes.CBC(iv))  # Create the cipher context
        cipher_context = cipher.encryptor()
        return cipher_context, iv

    def _write_encrypted_block(self, block_type: bytes, data: bytes):
        """
        Write a block of encrypted data.

        :param block_type: The four letter block type.
        :param data: The data of the block.
        """
        cipher_context, iv = self._prepare_encryption()
        encrypted_data = len(data).to_bytes(SIZE_VALUE_LENGTH, byteorder=SIZE_ENDIANNESS, signed=False)
        encrypted_data += iv
        padding = b''
        if (last_short_block := len(data) % AES_BLOCK_SIZE_BYTES) > 0:
            padding = os.urandom(AES_BLOCK_SIZE_BYTES - last_short_block)
        encrypted_data += cipher_context.update(data + padding)
        encrypted_data += cipher_context.finalize()
        self._write_block(block_type, encrypted_data)

    def _write_file_header(self):
        """
        Write the initial portion of the file.
        """
        self._write_magic()
        self._write_configuration()
        self._write_public_key_hash()
        self._create_and_write_encryption_key()

    def _write_meta_data(self, meta: Dict[str, Any]):
        """
        Write the meta data to the file.

        :param meta: The dictionary with the meta data.
        """
        if not meta:
            self._write_block(b'META', b'')
            self._write_block(b'MDHA', b'')
            return
        meta_data = json.dumps(meta).encode('utf-8')
        self._write_encrypted_block(b'META', meta_data)
        # Write the hash of the metadata.
        self._write_encrypted_block(b'MDHA', hashlib.sha3_512(meta_data).digest())

    def _write_bytes_data(self, source_data: bytes):
        """
        Write the bytes data encrypted to the destination

        :param source_data: The bytes object with the data to write.
        """
        source_size = len(source_data)
        if source_size == 0:
            # Write two empty blocks for empty files. This is safer than writing encryption data.
            self._write_block(b'DATA', b'')
            self._write_block(b'DTHA', b'')
            return
        self._write_encrypted_block(b'DATA', source_data)
        self._write_encrypted_block(b'DTHA', hashlib.sha3_512(source_data).digest())

    def _write_file_data(self, source_size: int, sf: io.BufferedIOBase):
        """
        Write the  file data encrypted to the destination.

        :param source_size: The size of the source file.
        :param sf: The open source file handle.
        """
        if source_size == 0:
            # Write two empty blocks for empty files. This is safer than writing encryption data.
            self._write_block(b'DATA', b'')
            self._write_block(b'DTHA', b'')
            return
        encryptor, iv = self._prepare_encryption()
        self._write_with_digest(b'DATA')  # Write the block type.
        # Calculate and write the size of the encrypted data.
        padding = b''
        if (last_short_block := source_size % AES_BLOCK_SIZE_BYTES) > 0:
            padding = os.urandom(AES_BLOCK_SIZE_BYTES - last_short_block)
        encrypted_data_size = SIZE_VALUE_LENGTH + len(iv) + source_size + len(padding)
        self._write_with_digest(
            encrypted_data_size.to_bytes(SIZE_VALUE_LENGTH, byteorder=SIZE_ENDIANNESS, signed=False))
        # Write the unencrypted size.
        self._write_with_digest(source_size.to_bytes(SIZE_VALUE_LENGTH, byteorder=SIZE_ENDIANNESS, signed=False))
        # Write the IV
        self._write_with_digest(iv)
        # Encrypt and write the actual data.
        block_hash_context = hashlib.sha3_512()
        while block := sf.read(WORKING_BLOCK_SIZE):
            block_hash_context.update(block)
            if len(block) < WORKING_BLOCK_SIZE:
                block += padding
            self._write_with_digest(encryptor.update(block))
        self._write_with_digest(encryptor.finalize())
        # Write the hash of the original data.
        self._write_encrypted_block(b'DTHA', block_hash_context.digest())

    def _write_data_chunk(self, data: bytes):
        """
        Write a single data chunk to the target stream

        :param data: The data to write.
        """
        self._write_with_digest(len(data).to_bytes(CHUNK_SIZE_LENGTH, byteorder=SIZE_ENDIANNESS, signed=False))
        if data:
            self._write_with_digest(data)

    def _write_chunked_data(self, data: bytes):
        """
        Write data in chunked format.

        :param data: The data block to write in the chunked format.
        """
        if not data:
            return
        self._chunked_buffer.extend(data)
        if len(self._chunked_buffer) >= MAXIMUM_CHUNK_SIZE:
            self._write_data_chunk(self._chunked_buffer[:MAXIMUM_CHUNK_SIZE])
            del self._chunked_buffer[:MAXIMUM_CHUNK_SIZE]

    def _flush_chunked_data(self):
        """
        Make sure any remaining chunked data is written.
        """
        if len(self._chunked_buffer) > 0:
            self._write_data_chunk(self._chunked_buffer)
        self._write_data_chunk(bytes())
        self._chunked_buffer.clear()

    def _stream_data(self, source_file_handle: io.BufferedIOBase):
        """
        Stream the encrypted data to the target stream.

        :param source_file_handle: The open source stream.
        """
        block = source_file_handle.read(WORKING_BLOCK_SIZE)
        if len(block) == 0:
            # Write two empty blocks for empty files. This is safer than writing encryption data.
            self._write_block(b'DATA', b'')
            self._write_block(b'DTHA', b'')
            return
        elif len(block) < WORKING_BLOCK_SIZE:
            self._write_encrypted_block(b'DATA', block)
            self._write_encrypted_block(b'DTHA', hashlib.sha3_512(block).digest())
            return
        encryptor, iv = self._prepare_encryption()
        self._write_chunked_block_header(b'DATA')
        # Write the IV
        self._write_chunked_data(iv)
        # Encrypt the first block
        block_hash_context = hashlib.sha3_512()
        block_hash_context.update(block)
        self._write_chunked_data(encryptor.update(block))
        # Encrypt all following blocks
        last_block = b''
        while block := source_file_handle.read(WORKING_BLOCK_SIZE):
            if last_block:
                self._write_chunked_data(encryptor.update(last_block))
            block_hash_context.update(block)
            last_block = block
        # At this point, we always get the last read block which was not encrypted yet.
        # Always apply ISO/IEC 9797-1 padding method 2 to the last block (even it is empty)
        last_block += b'\x80'
        misalignment = len(last_block) % AES_BLOCK_SIZE_BYTES
        if misalignment:
            last_block += bytes(AES_BLOCK_SIZE_BYTES - misalignment)
        self._write_chunked_data(encryptor.update(last_block))
        self._write_chunked_data(encryptor.finalize())
        # End the chunked data stream
        self._flush_chunked_data()
        # Write the hash of the original data.
        self._write_encrypted_block(b'DTHA', block_hash_context.digest())

    def _write_end_with_hash(self):
        """
        Write the end mark with the file hash.
        """
        self.destination_file_handle.write(b'ENDH')
        file_digest = self.destination_file_digest.digest()
        self.destination_file_handle.write(
            len(file_digest).to_bytes(SIZE_VALUE_LENGTH, byteorder=SIZE_ENDIANNESS, signed=False))
        self.destination_file_handle.write(file_digest)
        self.destination_file_handle.flush()

    def _clean_up(self):
        """
        Remove information not needed after encryption.
        """
        self.destination_file_handle = None
        self.destination_file_digest = None
        self.algorithm = None
        self._chunked_buffer.clear()

    @staticmethod
    def _verify_metadata(meta: Dict[str, Any]):
        """
        Check if the metadata has the right format.

        :param meta: An existing metadata dictionary or `None`.
        """
        if meta:
            if not isinstance(meta, dict):
                raise ValueError('Metadata has to be a dictionary.')
            for key in meta.keys():
                if not isinstance(key, str):
                    raise ValueError('Metadata keys have to be strings.')

    @staticmethod
    def _add_source_metadata(source: Path, meta: Dict[str, Any]) -> Dict[str, Any]:
        """
        Add the metadata from the source file.

        :param source: The source file.
        :param meta: The existing metadata.
        :return: The new metadata.
        """
        if not meta:
            meta = {}
        else:
            meta = meta.copy()
        if 'file_path' not in meta:
            meta['file_path'] = str(source.absolute())
        if 'file_name' not in meta:
            meta['file_name'] = str(source.name)
        if 'file_size' not in meta:
            meta['file_size'] = source.stat().st_size
        if 'created' not in meta:
            meta['created'] = datetime.utcfromtimestamp(source.stat().st_birthtime).isoformat()
        if 'modified' not in meta:
            meta['modified'] = datetime.utcfromtimestamp(source.stat().st_mtime).isoformat()
        return meta

    def save_encrypted(self, source_data: bytes, destination: Path, meta: Dict[str, Any] = None):
        """
        Save small (!) amounts of data encrypted at the destination.

        :param source_data: The data to encrypt and save.
        :param destination: The path to the destination file.
            Use the suffix `.fonfenc` to the target files for best compatibility.
        :param meta: A dictionary with metadata for this file.
        """
        if not isinstance(source_data, bytes):
            raise ValueError('`source_data` has to be a bytes object.')
        if not isinstance(destination, Path):
            raise ValueError('`destination` has to be a `Path` from pathlib.')
        self._verify_metadata(meta)
        with destination.open('wb') as destination_file_handle:
            self.destination_file_digest = hashlib.sha3_512()
            self.destination_file_handle = destination_file_handle
            self._write_file_header()
            self._write_meta_data(meta)
            self._write_bytes_data(source_data)
            self._write_end_with_hash()
        self._clean_up()

    def copy_encrypted(self, source: Path, destination: Path, meta: Dict[str, Any] = None,
                       add_source_metadata=False):
        """
        Copy a file, and store it encrypted at the destination.

        :param source: The path to the source file.
        :param destination: The path to the destination file.
            Use the suffix `.fonfenc` to the target files for best compatibility.
        :param meta: A dictionary with metadata for this file.
        :param add_source_metadata: Include the metadata for the given source file.
            This includes the fields `file_path`, `file_name`, `file_size`, `created`, `modified`.
            Only missing fields are added to the metadata.
        :raises DataTooLargeError: If the source file exceeds the maximum file size limit of 1 TB.
        """
        if not isinstance(source, Path):
            raise ValueError('`source` has to be a `Path` from pathlib.')
        if not isinstance(destination, Path):
            raise ValueError('`destination` has to be a `Path` from pathlib.')
        if not isinstance(add_source_metadata, bool):
            raise ValueError('`add_source_metadata` has to be a boolean value.')
        self._verify_metadata(meta)
        source_file_size = source.stat().st_size
        if source_file_size > FILE_SIZE_LIMIT:
            raise DataTooLargeError('File sizes larger than 1TB are not supported.')
        with source.open('rb') as source_file_handle, destination.open('wb') as destination_file_handle:
            self.destination_file_digest = hashlib.sha3_512()
            self.destination_file_handle = destination_file_handle
            self._write_file_header()
            if add_source_metadata:
                meta = self._add_source_metadata(source, meta)
            self._write_meta_data(meta)
            self._write_file_data(source_file_size, source_file_handle)
            self._write_end_with_hash()
        self._clean_up()

    def stream_encrypted(self, source_io: io.BufferedIOBase, destination_io: io.BufferedIOBase,
                         meta: Dict[str, Any] = None):
        """
        Read data from a stream and write it encrypted into another stream.

        For short streams, smaller than `WORKING_BLOCK_SIZE` this will write the destination stream
        on the fly. Larger streams will write some dummy data block header, encrypt all the data from the
        source stream, but then **seek** back to the data block start to write the correctly original
        and encrypted sizes.

        :param source_io: The **open** source stream, compatible with io.BufferedIOBase.
        :param destination_io: The **open** destination stream, compatible with io.BufferedIOBase.
        :param meta: A dictionary with metadata for this file.
        """
        if not isinstance(source_io, io.BufferedIOBase):
            raise ValueError('`source_io` has to be a subclass of `io.BufferedIOBase`.')
        if not source_io.readable():
            raise ValueError('The source stream has to be readable.')
        if not isinstance(destination_io, io.BufferedIOBase):
            raise ValueError('`destination_io` has to be a subclass of `io.BufferedIOBase`.')
        if not destination_io.writable():
            raise ValueError('The destination stream has to be writeable.')
        if not destination_io.seekable():
            raise ValueError('The destination stream has to be seekable.')
        self._verify_metadata(meta)
        self.destination_file_digest = hashlib.sha3_512()
        self.destination_file_handle = destination_io
        self._write_file_header()
        self._write_meta_data(meta)
        self._stream_data(source_io)
        self._write_end_with_hash()
        self._clean_up()



# Copyright 2021 by Tobias Erbsland / EducateIT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module contains the shared values for the encryption, decryption and key generation.
Do not change these values!
"""

RSA_KEY_SIZE = 4096  # The used key size. Do not change this value!

AES_KEY_LENGTH_BITS = 256  # The used key length. Do not change this value!
AES_KEY_LENGTH_BYTES = AES_KEY_LENGTH_BITS // 8  # The key length on bytes
AES_BLOCK_SIZE_BITS = 128  # The block size for the encryption. Do not change this value!
AES_BLOCK_SIZE_BYTES = AES_BLOCK_SIZE_BITS // 8  # The block size in bytes.
AES_IV_LENGTH_BYTES = 16  # The length of the IV for the cipher in bytes.

WORKING_BLOCK_SIZE = 4096  # The block size to read/write data from a file. Has to align with AES block size!
MAXIMUM_CHUNK_SIZE = 0xFFFF  # The maximum size of a block in the chunked format.
SIZE_VALUE_LENGTH = 8  # Use 8 bytes to store the size values.
CHUNK_SIZE_LENGTH = 2  # Use 2 bytes to store the chunk size values.
SIZE_ENDIANNESS = 'big'  # The endianness of the size values.

MAXIMUM_BLOCK_SIZE_VALUE = 0xffff_0000_0000_0000  # The maximum valid block size
CHUNKED_BLOCK_SIZE_VALUE = 0xffff_8000_0000_0000  # The special value indicating a block uses chunked data.

ENCRYPTION_DATA_GAIN = SIZE_VALUE_LENGTH + AES_IV_LENGTH_BYTES + AES_BLOCK_SIZE_BYTES

FILE_MAGIC = b'\xfeFFE\x0d\x0a\x1a\x0a'
FILE_CONFIG_TEXT = b'k:RSA-4096,e:AES-256,b:CBC,h:SHA3-512,v:1'
FILE_SIZE_LIMIT = 10_000_000_000_000

KNOWN_BLOCK_TYPES = [  # All known block types.
    b'CONF',
    b'EPUB',
    b'ESYM',
    b'META',
    b'MDHA',
    b'DATA',
    b'DTHA',
    b'ENDH']
